{$B-}
(* This mode (skmhltxt) created by Michail A.Baikov
   Copyright (c) 1998-1999 by Sergey Korowkin
   Portions Copyright (pc) 1999 by Michail A.Baikov
*)

unit skMHLtxt;

interface
uses
     skMHL, skCommon;

const
 fmbWrongPath           = $1000;
 fmbMessageNotFound     = $1001;
 fmbCannotCreateStream  = $1002;
 fmbAlreadyClosed       = $1003;
 fmbBufferOverflow      = $1004;

 MaxMapSize             = $F000;

type

 TFidoMap = array[0..MaxMapSize div 8 - 1] of Byte;

 PTextMessageBase = ^TTextMessageBase;
 TTextMessageBase = object(TMessageBase)
  constructor Init;
  destructor Done; virtual;
  function Open(const Path: String): Boolean; virtual;
  function Create(const Path: String): Boolean; virtual;
  procedure Close; virtual;
  function Exists(Message: Longint): Boolean; virtual;
  function Current: Longint; virtual;
  function GetLocation: Longint; virtual;
  procedure SetLocation(Location: Longint); virtual;
  function OpenMessage: Boolean; virtual;
  function CloseMessage: Boolean; virtual;
  function GetHighest: Longint; virtual;
  function GetCount: Longint; virtual;
  function GetFrom: String; virtual;
  function GetTo: String; virtual;
  function GetSubject: String; virtual;
  procedure GetFromAddress(var Address: TAddress); virtual;
  procedure GetToAddress(var Address: TAddress); virtual;
  procedure GetFromAndToAddress(var FromAddress, ToAddress: TAddress); virtual;
  procedure SetFrom(const S: String); virtual;
  procedure SetTo(const S: String); virtual;
  procedure SetSubject(const S: String); virtual;
  procedure SetFromAddress(var Address: TAddress; const FreshMSGID: Boolean); virtual;
  procedure SetToAddress(var Address: TAddress); virtual;
  procedure SetFromAndToAddress(var FromAddress, ToAddress: TAddress); virtual;
  function GetAttribute(Attribute: Longint): Boolean; virtual;
  procedure SetAttribute(Attribute: Longint; Enable: Boolean); virtual;
  procedure GetWrittenDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure GetArrivedDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure SetWrittenDateTime(var DateTime: TMessageBaseDateTime); virtual;
  procedure SetArrivedDateTime(var DateTime: TMessageBaseDateTime); virtual;
  function WriteMessage: Boolean; virtual;
  function CreateNewMessage: Boolean; virtual;
  function KillMessage: Boolean; virtual;
 private
  Directory: String;
  TotalMessages: Longint;
  Map: TFidoMap;
  HighMessage: Longint;
  Link: PMessageBaseStream;
  procedure RebuildIndex;
  procedure SetMapDot(Num: Integer; Value: Boolean);
  function GetMapDot(Num: Integer): Boolean;
  function MapAttribute(var Attribute: Longint): Boolean;
 end;

implementation

const
 MapPos: Array[0..7] Of Byte = (1, 2, 4, 8, 16, 32, 64, 128);

{* TTextMessageBase *}

constructor TTextMessageBase.Init;
 begin
  inherited Init;
  Link:=Nil;
 end;

destructor TTextMessageBase.Done;
 begin
  CloseMessage;
  Close;
  inherited Done;
 end;

function TTextMessageBase.Open(const Path: String): Boolean;
 begin
  Open:=False;
  Opened:=False;
  Directory:=Path;
  if Directory[Length(Directory)] <> '\' then Directory:=Concat(Directory, '\');
  if not DirectoryExists(Path) then
   begin
    Status:=fmbWrongPath;
    Exit;
   end;
  RebuildIndex;
  Open:=True;
  Opened:=True;
 end;

function TTextMessageBase.Create(const Path: String): Boolean;
 begin
  Create:=False;
  Opened:=False;
  Directory:=Path;
  if Directory[Length(Directory)] <> '\' then Directory:=Concat(Directory, '\');
  if not CreateDirectory(Directory) then Exit;
  RebuildIndex;
  Create:=True;
  Opened:=True;
 end;

procedure TTextMessageBase.Close;
 begin
  Directory:='';
  Opened:=False;
 end;

function TTextMessageBase.Exists(Message: Longint): Boolean;
 begin
  Exists:=GetMapDot(Message);
 end;

function TTextMessageBase.Current: Longint;
 begin
  Current:=CurrentMessage;
 end;

function TTextMessageBase.GetLocation: Longint;
 begin
  GetLocation:=CurrentMessage;
 end;

procedure TTextMessageBase.SetLocation(Location: Longint);
 begin
  CurrentMessage:=Location;
 end;

function TTextMessageBase.OpenMessage: Boolean;
 begin
  OpenMessage:=False;
  if not Exists(CurrentMessage) then
   begin
    Status:=fmbMessageNotFound;
    Exit;
   end;
  if Link <> Nil then CloseMessage;
  Link:=CreateMessageBaseFileStream(Directory + LongToStr(CurrentMessage) + '.DSK', smOpen);
  if Link^.Status <> smOk then
   begin
    Dispose(Link, Done);
    Link:=Nil;
    Status:=fmbCannotCreateStream;
    Exit;
   end;
  MessageText:=CreateMessageBaseMemoryStream(MaxMessageSize);
  if MessageText^.Status <> smOk then
   begin
    Dispose(Link, Done);
    Dispose(MessageText, Done);
    Link:=Nil;
    Status:=fmbCannotCreateStream;
    Exit;
   end;
  Link^.Seek(0);
//  Link^.Read(Header, SizeOf(Header));
  MessageText^.CopyFrom(Link^, Link^.GetSize - Link^.GetPos - 1);
  MessageText^.Seek(0);
  OpenMessage:=True;
 end;

function TTextMessageBase.CloseMessage: Boolean;
 begin
  CloseMessage:=False;
  if Link = Nil then
   begin
    Status:=fmbAlreadyClosed;
    Exit;
   end;
  Dispose(Link, Done);
  Dispose(MessageText, Done);
  Link:=Nil;
  CloseMessage:=True;
 end;

function TTextMessageBase.GetHighest: Longint;
 begin
  GetHighest:=HighMessage;
 end;

function TTextMessageBase.GetCount: Longint;
 begin
  GetCount:=TotalMessages;
 end;

function TTextMessageBase.GetFrom: String;
 begin
  GetFrom:=''; // FromASCIIZ(@Header.FromUser);
 end;

function TTextMessageBase.GetTo: String;
 begin
  GetTo:=''; // FromASCIIZ(@Header.ToUser);
 end;

function TTextMessageBase.GetSubject: String;
 begin
  GetSubject:=''; // FromASCIIZ(@Header.Subject);
 end;

procedure TTextMessageBase.GetFromAddress(var Address: TAddress);
 begin
  Address.Zone:=0;
  Address.Net:=0; // Header.OrigNet;
  Address.Node:=0; // Header.OrigNode;
  Address.Point:=0;

// inherited GetFromAddress(Address);
 end;

{//mab}
procedure TTextMessageBase.GetToAddress(var Address: TAddress);
 var
  TempAddress: TAddress;
 begin
  Address.Zone:=0;
  Address.Point:=0;

//  inherited GetToAddress(Address);

  Address.Net:=0; // Header.DestNet;
  Address.Node:=0; // Header.DestNode;

(*   if Address.Zone = 0 then
   begin
    GetFromAddress(TempAddress);
    Address.Zone:=TempAddress.Zone;
   end; *)
 end;
{//end}

procedure TTextMessageBase.GetFromAndToAddress(var FromAddress, ToAddress: TAddress);
 begin
  FromAddress.Zone:=0;
  FromAddress.Net:=0; // Header.OrigNet;
  FromAddress.Node:=0; // Header.OrigNode;
  FromAddress.Point:=0;
  ToAddress.Zone:=0;
  ToAddress.Net:=0; // Header.DestNet;
  ToAddress.Node:=0; // Header.DestNode;
  ToAddress.Point:=0;
//  inherited GetFromAndToAddress(FromAddress, ToAddress);
 end;

procedure TTextMessageBase.SetFrom(const S: String);
 begin
//  ToASCIIZ(S, @Header.FromUser);
 end;

procedure TTextMessageBase.SetTo(const S: String);
 begin
//  ToASCIIZ(S, @Header.ToUser);
 end;

procedure TTextMessageBase.SetSubject(const S: String);
 begin
  // ToASCIIZ(S, @Header.Subject);
 end;

procedure TTextMessageBase.SetFromAddress(var Address: TAddress; const FreshMSGID: Boolean);
 begin
//  inherited SetFromAddress(Address, FreshMSGID);
//  Header.OrigNet:=0; // Address.Net;
//  Header.OrigNode:=0; // Address.Node;
 end;

procedure TTextMessageBase.SetToAddress(var Address: TAddress);
 begin
//  inherited SetToAddress(Address);
//  Header.DestNet:=0; // Address.Net;
//  Header.DestNode:=0; // Address.Node;
 end;

procedure TTextMessageBase.SetFromAndToAddress(var FromAddress, ToAddress: TAddress);
 begin
//  inherited SetFromAndToAddress(FromAddress, ToAddress);
//  Header.OrigNet:=0; // FromAddress.Net;
//  Header.OrigNode:=0; // FromAddress.Node;
//  Header.DestNet:=0; // ToAddress.Net;
//  Header.DestNode:=0; // ToAddress.Node;
 end;

function TTextMessageBase.GetAttribute(Attribute: Longint): Boolean;
 begin
  if not MapAttribute(Attribute) then
   begin
    GetAttribute:=False;
    Exit;
   end;
  GetAttribute:=False; // Header.Attr and Attribute = Attribute;
 end;

procedure TTextMessageBase.SetAttribute(Attribute: Longint; Enable: Boolean);
 begin
  if not MapAttribute(Attribute) then Exit;
//  if Enable then
//   Header.Attr:=Header.Attr or Attribute
//  else
//   Header.Attr:=Header.Attr and not Attribute;
 end;

{ Warning!
  TFidoMessageBase.GetWrittenDateTime was tested only with format of
  keeping of date named 'Fido'. Work with other formats of keeping of
  date is not guaranteed.
}
procedure TTextMessageBase.GetWrittenDateTime(var DateTime: TMessageBaseDateTime);
 var
  S: String;
  K: Byte;
 begin
(*  FillChar(DateTime, SizeOf(DateTime), 0);
  S:=FromASCIIZ(@Header.DateTime);
  if S[3] = ' ' then
   if S[11] = ' ' then {Fido - "DD MMM YY  HH:MM:SS"}
    begin
     StrToWord(ExtractWord(1, S, [' ']), DateTime.Day);
     MonthStringToMonthNumber(ExtractWord(2, S, [' ']), DateTime.Month);
     StrToWord(ExtractWord(3, S, [' ']), DateTime.Year);
     Delete(S, 1, 11);
     StrToWord(ExtractWord(1, S, [':']), DateTime.Hour);
     StrToWord(ExtractWord(2, S, [':']), DateTime.Min);
     StrToWord(ExtractWord(3, S, [':', ' ']), DateTime.Sec);
    end
   else {Opus - "DD MMM YY HH:MM:SS"}
    begin
     StrToWord(ExtractWord(1, S, [' ']), DateTime.Day);
     MonthStringToMonthNumber(ExtractWord(2, S, [' ']), DateTime.Month);
     StrToWord(ExtractWord(3, S, [' ']), DateTime.Year);
     Delete(S, 1, 10);
     StrToWord(ExtractWord(1, S, [':']), DateTime.Hour);
     StrToWord(ExtractWord(2, S, [':']), DateTime.Min);
     StrToWord(ExtractWord(3, S, [':', ' ']), DateTime.Sec);
    end
  else
   if S[4] = ' ' then {SeaDog - "DOW DD MMM YY HH:MM"}
    begin
     Delete(S, 1, 4);
     StrToWord(ExtractWord(1, S, [' ']), DateTime.Day);
     MonthStringToMonthNumber(ExtractWord(2, S, [' ']), DateTime.Month);
     StrToWord(ExtractWord(3, S, [' ']), DateTime.Year);
     Delete(S, 1, 10);
     StrToWord(ExtractWord(1, S, [':']), DateTime.Hour);
     StrToWord(ExtractWord(2, S, [':']), DateTime.Min);
     DateTime.Sec:=0;
    end
   else
    if S[3] = '-' then {Wierd - "DD-MM-YYYY HH:MM:SS"}
     begin
      for K:=1 to Length(S) do
       if S[K] in ['-', ':'] then S[K]:=' ';
      StrToWord(ExtractWord(1, S, [' ']), DateTime.Day);
      StrToWord(ExtractWord(2, S, [' ']), DateTime.Month);
      StrToWord(ExtractWord(3, S, [' ']), DateTime.Year);
      StrToWord(ExtractWord(4, S, [' ']), DateTime.Hour);
      StrToWord(ExtractWord(5, S, [' ']), DateTime.Min);
      StrToWord(ExtractWord(6, S, [' ']), DateTime.Sec);
     end;
  if DateTime.Year < 1900 then
   Inc(DateTime.Year, 1900);
{ I don't know how to parse binary format of keeping of date, so this call
  was disabled:
  ???DateTimeToMessageBaseDateTime(Header.DateWritten, DateTime);} *)
 end;

procedure TTextMessageBase.GetArrivedDateTime(var DateTime: TMessageBaseDateTime);
 begin
{ I don't know how to parse binary format of keeping of date, so this call
  was disabled:
  ???DateTimeToMessageBaseDateTime(Header.DateArrived, DateTime);}
 end;

procedure TTextMessageBase.SetWrittenDateTime(var DateTime: TMessageBaseDateTime);
 function LeadingZero(Number: Word): String;
  var
   S: String[2];
  begin
{y2k by mab}
  if number>=100 then dec(number,100);
{y2k by mab}
   Str(Number, S);
   if S[0] = #1 then
    S:=Concat('0', S);
   LeadingZero:=S;
  end;
 begin
(*   ToASCIIZ(LeadingZero(DateTime.Day) + ' ' +
           MonthNumberToMonthString(DateTime.Month) + ' ' +
           LeadingZero(DateTime.Year - 1900) + '  ' +
           LeadingZero(DateTime.Hour) + ':' +
           LeadingZero(DateTime.Min) + ':' +
           LeadingZero(DateTime.Sec), @Header.DateTime); *)
{ I don't know how to parse binary format of keeping of date, so this call
  was disabled:
  MessageBaseDateTimeTo???DateTime(DateTime, Header.DateWritten);}
 end;

procedure TTextMessageBase.SetArrivedDateTime(var DateTime: TMessageBaseDateTime);
 begin
{ I don't know how to parse binary format of keeping of date, so this call
  was disabled:
  MessageBaseDateTimeTo???DateTime(DateTime, Header.DateArrived);}
 end;

function TTextMessageBase.WriteMessage: Boolean;
 var
  B: System.Byte;
 begin
  WriteMessage:=True;
  Link^.Reset;
  Link^.Seek(0);
//  Link^.Write(Header, SizeOf(Header));
  MessageText^.Seek(0);
  Link^.CopyFrom(MessageText^, MessageText^.GetSize);
  B:=0;
  Link^.Write(B, SizeOf(B));
  Link^.Truncate;
  Link^.Flush;
 end;

function TTextMessageBase.CreateNewMessage: Boolean;
 begin
  CreateNewMessage:=False;
  CurrentMessage:=GetHighest + 1;
  Inc(HighMessage);
  SetMapDot(CurrentMessage, True);
  if Link <> Nil then CloseMessage;
  Link:=CreateMessageBaseFileStream(Directory + LongToStr(CurrentMessage) + '.DSK', smCreate);
  if Link^.Status <> smOk then
   begin
    Dispose(Link, Done);
    Link:=Nil;
    Exit;
   end;
  MessageText:=CreateMessageBaseMemoryStream(MaxMessageSize);
  if MessageText^.Status <> smOk then
   begin
    Dispose(Link, Done);
    Dispose(MessageText, Done);
    Link:=Nil;
    Exit;
   end;
//  FillChar(Header, SizeOf(Header), 0);
  CreateNewMessage:=True;
 end;

function TTextMessageBase.KillMessage: Boolean;
 begin
  KillMessage:=False;
  if not Exists(CurrentMessage) then Exit;
  if not KillFile(Directory + LongToStr(CurrentMessage) + '.TXT') then Exit;
  SetMapDot(CurrentMessage, False);
  KillMessage:=True;
 end;

{* Private methods *}

procedure TTextMessageBase.RebuildIndex;
 var
  Find: PMessageBaseFind;
{$IFDEF VIRTUALPASCAL}
  Num, Code: Longint;
{$ELSE}
  Num, Code: Integer;
{$ENDIF}
 begin
//  FillChar(Map, SizeOf(Map), 0);
  TotalMessages:=0;
  HighMessage:=0;
  Find:=CreateMessageBaseFind {DefaultFindCreator}; {mab};
  if Find^.StartSearch(Directory + '*.TXT', faReadonly + faArchive) then
   repeat
    Val(Copy(Find^.iName, 1, Pos('.', Find^.iName) - 1), Num, Code);
    if (Code = 0) and (Num > 0) then
     begin
      SetMapDot(Num, True);
      Inc(TotalMessages);
      if Num > HighMessage then
       HighMessage:=Num;
     end;
   until not Find^.NextSearch;
  Find^.StopSearch;
  Dispose(Find, Done);
 end;

procedure TTextMessageBase.SetMapDot(Num: Integer; Value: Boolean);
 var
  N: Integer;
 begin
  if Num >= MaxMapSize then Exit;
  N:=Num shr $03;
  if Value then
   Map[N]:=Map[N] or MapPos[Num and $07]
  else
   Map[N]:=Map[N] and not MapPos[Num and $07];
 end;

function TTextMessageBase.GetMapDot(Num: Integer): Boolean;
 begin
  if Num >= MaxMapSize then
   begin
    GetMapDot:=False;
    Exit;
   end;
  GetMapDot:=Map[Num shr $03] and MapPos[Num and $07] <> 0;
 end;

function TTextMessageBase.MapAttribute(var Attribute: Longint): Boolean;
 begin
  MapAttribute:=True;
 end;

end.
